// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "shader.h"

CShader base_shader;      // zakladny shader
CShader light_shader;     // shader pre osvetlenie
CShader brick_shader;     // shader - tehly
CShader toon_shader;      // shader - rozpravka
CShader wave_shader;      // shader - vlna
CShader wood_shader;      // shader - drevo
CShader nm_shader;        // shader - normalmap
CShader px_shader;        // shader - parallax
CShader water_shader;     // shader - water
CDOFshaders dof_shader;   // shader - Depth Of Field
CShader chromatic_shader; // shader - Chromatic Dispersion
CShader bloom_shader;     // shader - bloom
CShader blur_shader;      // shader - blur
CShader hdr_shader;       // shader - High Dynamic Range
CShader ssao_shader[5];   // shader - Screen Space Ambient Occlusion
CShader shadow_shader;    // shader - shadow
// pre deferred shading
CShader raw_shader;       // nespracovane udaje
CShader def_shader;       // vypocet osvetlenia
CShader msaa_shader;      // multisample antialiasing

// nacitanie obsahu textoveho suboru ***************************************
char *TextFileRead(const char *FNAME)
{
 FILE *f_p=NULL;
 char *f_content=NULL;
 int f_size=0;

 // ziskanie velkosti suboru
 f_p=fopen(FNAME,"rb");
 if (f_p==NULL) { printf("Nemozno otvorit subor: %s!\n",FNAME); return(NULL); }
 fseek(f_p,0,SEEK_END);
 f_size=ftell(f_p);
 fclose(f_p);

 // ak subor nie je prazdny
 if (f_size>0) {
  f_p=fopen(FNAME,"rt");
  // nacitanie obsahu textoveho suboru
  f_content=new char[f_size+1];
  if (f_content==NULL) { printf("Nemozno vyhradit pamat!\n"); fclose(f_p); return(NULL); }
  f_size=(int) fread(f_content,sizeof(char),f_size,f_p);
  f_content[f_size]='\0';
  fclose(f_p);
 }
 else { printf("Subor je prazdny!\n"); return(NULL); }

 return(f_content);
}

// konstruktor *************************************************************
CShader::CShader()
{
 vertex=0; fragment=0; program=0;
}

// destruktor **************************************************************
CShader::~CShader()
{
#ifdef OGL20
 if (vertex && fragment && program) {
  // odpojenie shaderov
  glDetachShader(program,vertex); glDetachShader(program,fragment);
  // zrusenie shader objektov
  glDeleteShader(vertex); glDeleteShader(fragment);
  // zrusenie program objektu
  glDeleteProgram(program);
 }
#else
 if (vertex && fragment && program) {
  glDetachObjectARB(program,vertex); glDetachObjectARB(program,fragment);
  glDeleteObjectARB(vertex); glDeleteObjectARB(fragment);
  glDeleteObjectARB(program);
 }
#endif
 vertex=0; fragment=0; program=0;
}

// inicializacia pouzitia shaderov *****************************************
int CShader::init(void)
{
 printf("GLSL init...\n");

#ifdef OGL20
 if (glewIsSupported("GL_VERSION_2_0")) printf("OpenGL 2.0 supported\n");
 else { printf("OpenGL 2.0 not supported\n"); return(1); }
#else
 if (GLEW_ARB_vertex_shader && GLEW_ARB_fragment_shader) printf("ARB extensions supported\n");
 else { printf("ARB extensions not supported\n"); return(1); }
#endif

 return(0);
}

// nacitanie kodu GLSL vertex & fragment shadera ***************************
int CShader::load(const char *VP,const char *FP)
{
 strcpy(vp_name,VP); // nazov vertex programu
 strcpy(fp_name,FP); // nazov fragment programu

 char *vp_src=TextFileRead(VP); // kod vertex programu
 char *fp_src=TextFileRead(FP); // kod fragment programu

 if (vp_src==NULL || fp_src==NULL) {
  printf("Nemozno nacitat kod GLSL shaderov!\n");
  DELETEA(vp_src); DELETEA(fp_src);
  return(1);
 }

 // nazov suborov s kodom vertex a fragment shadera
 printf("%s %s\n",VP,FP);

#ifdef OGL20
 // vytvorenie shader objektu (kontajnera)
 vertex=glCreateShader(GL_VERTEX_SHADER);
 fragment=glCreateShader(GL_FRAGMENT_SHADER);
 // kod shaderov
 glShaderSource(vertex,1,(const char**) &vp_src,NULL);
 glShaderSource(fragment,1,(const char**) &fp_src,NULL);
 // kompilacia shaderov
 glCompileShader(vertex);
 glGetShaderiv(vertex,GL_INFO_LOG_LENGTH,&log_length);
 if (log_length>0 && log_length<1000) {
  glGetShaderInfoLog(vertex,log_length,NULL,log);
  printf("vertex_shader_log: %s",log);
 }
 glCompileShader(fragment);
 glGetShaderiv(fragment,GL_INFO_LOG_LENGTH,&log_length);
 if (log_length>0 && log_length<1000) {
  glGetShaderInfoLog(fragment,log_length,NULL,log);
  printf("fragment_shader_log: %s",log);
 }
 // vytvorenie program objektu (kontajnera)
 program=glCreateProgram();
 // pripojenie shaderov
 glAttachShader(program,vertex);
 glAttachShader(program,fragment);
 // zlozenie programu
 glLinkProgram(program);
 glGetProgramiv(program,GL_INFO_LOG_LENGTH,&log_length);
 if (log_length>0 && log_length<1000) {
  glGetProgramInfoLog(program,log_length,NULL,log);
  printf("program_log: %s\n",log);
 }
#else
 vertex=glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
 fragment=glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);

 glShaderSourceARB(vertex,1,(const char**) &vp_src,NULL);
 glShaderSourceARB(fragment,1,(const char**) &fp_src,NULL);

 glCompileShaderARB(vertex);
 glGetObjectParameterivARB(vertex,GL_OBJECT_INFO_LOG_LENGTH_ARB,&log_length);
 if (log_length>0 && log_length<1000) {
  glGetInfoLogARB(vertex,log_length,NULL,log);
  printf("vertex_shader_log: %s\n",log);
 }
 glCompileShaderARB(fragment);
 glGetObjectParameterivARB(fragment,GL_OBJECT_INFO_LOG_LENGTH_ARB,&log_length);
 if (log_length>0 && log_length<1000) {
  glGetInfoLogARB(fragment,log_length,NULL,log);
  printf("fragment_shader_log: %s\n",log);
 }

 program=glCreateProgramObjectARB();

 glAttachObjectARB(program,vertex);
 glAttachObjectARB(program,fragment);

 glLinkProgramARB(program);
 glGetObjectParameterivARB(program,GL_OBJECT_INFO_LOG_LENGTH_ARB,&log_length);
 if (log_length>0 && log_length<1000) {
  glGetInfoLogARB(program,log_length,NULL,log);
  printf("program_log: %s\n",log);
 }
#endif

 // uvolnenie pamate vyhradenej pre kod shaderov
 DELETEA(vp_src); DELETEA(fp_src);

 return(0);
}

// nacitanie kodu GLSL vertex & fragment shadera ***************************
int CShader::reload(void)
{
#ifdef OGL20
 if (vertex && fragment && program) {
  // odpojenie shaderov
  glDetachShader(program,vertex); glDetachShader(program,fragment);
  // zrusenie shader objektov
  glDeleteShader(vertex); glDeleteShader(fragment);
  // zrusenie program objektu
  glDeleteProgram(program);
 }
#else
 if (vertex && fragment && program) {
  glDetachObjectARB(program,vertex); glDetachObjectARB(program,fragment);
  glDeleteObjectARB(vertex); glDeleteObjectARB(fragment);
  glDeleteObjectARB(program);
 }
#endif

 return(load(vp_name,fp_name));
}

// zapnutie GLSL shadera ***************************************************
void CShader::on(void)
{
#ifdef OGL20
 // pouzitie programu
 glUseProgram(program);
#else
 glUseProgramObjectARB(program);
#endif
}

// vypnutie GLSL shadera ***************************************************
void CShader::off(void)
{
#ifdef OGL20
 // nepouzitie programu
 glUseProgram(0);
#else
 glUseProgramObjectARB(0);
#endif
}

// nastavenie atributov ****************************************************
// POZOR!!! nastavenie uniform premennych moze byt vykonane az po vykonani
// vsetkych predoslych bindattriblocation
void CShader::bindattriblocation(const GLuint I,const char *NAME)
{
#ifdef OGL20
 // pripojenie atributu
 // atribut NAME programu program sa spoji s generickym atributom I
 // program musi byt znovu linkovany
 glBindAttribLocation(program,I,NAME);
 glLinkProgram(program);
#else
 glBindAttribLocationARB(program,I,NAME);
 glLinkProgramARB(program);
#endif
}

// nastavenie uniform premennych *******************************************
void CShader::uniformI(const char *NAME,const int I)
{
#ifdef OGL20
 // umiestnenie premennej
 loc=glGetUniformLocation(program,NAME);
 // nastavenie premennej
 if (loc!=-1) glUniform1i(loc,I);
#else
 loc=glGetUniformLocationARB(program,NAME);
 if (loc!=-1) glUniform1iARB(loc,I);
#endif
}
void CShader::uniform1F(const char *NAME,const float F)
{
#ifdef OGL20
 // umiestnenie premennej
 loc=glGetUniformLocation(program,NAME);
 // nastavenie premennej
 if (loc!=-1) glUniform1f(loc,F);
#else
 loc=glGetUniformLocationARB(program,NAME);
 if (loc!=-1) glUniform1fARB(loc,F);
#endif
}
void CShader::uniform2F(const char *NAME,const float F0,const float F1)
{
#ifdef OGL20
 // umiestnenie premennej
 loc=glGetUniformLocation(program,NAME);
 // nastavenie premennej
 if (loc!=-1) glUniform2f(loc,F0,F1);
#else
 loc=glGetUniformLocationARB(program,NAME);
 if (loc!=-1) glUniform2fARB(loc,F0,F1);
#endif
}
void CShader::uniform3F(const char *NAME,const float F0,const float F1,const float F2)
{
#ifdef OGL20
 // umiestnenie premennej
 loc=glGetUniformLocation(program,NAME);
 // nastavenie premennej
 if (loc!=-1) glUniform3f(loc,F0,F1,F2);
#else
 loc=glGetUniformLocationARB(program,NAME);
 if (loc!=-1) glUniform3fARB(loc,F0,F1,F2);
#endif
}
void CShader::uniformFV(const char *NAME,const int C,const GLfloat *V)
{
#ifdef OGL20
 // umiestnenie premennej
 loc=glGetUniformLocation(program,NAME);
 // nastavenie premennej
 if (loc!=-1) glUniform3fv(loc,C,V);
#else
 loc=glGetUniformLocationARB(program,NAME);
 if (loc!=-1) glUniform3fvARB(loc,C,V);
#endif
}
void CShader::uniform4F(const char *NAME,const float F0,const float F1,const float F2,const float F3)
{
#ifdef OGL20
 // umiestnenie premennej
 loc=glGetUniformLocation(program,NAME);
 // nastavenie premennej
 if (loc!=-1) glUniform4f(loc,F0,F1,F2,F3);
#else
 loc=glGetUniformLocationARB(program,NAME);
 if (loc!=-1) glUniform4fARB(loc,F0,F1,F2,F3);
#endif
}
void CShader::uniformmatrixFV(const char *NAME,const int C,const GLfloat *M)
{
#ifdef OGL20
 // umiestnenie matice
 loc=glGetUniformLocation(program,NAME);
 // nastavenie matice
 if (loc!=-1) glUniformMatrix4fv(loc,C,GL_FALSE,M); // umiestnenie, C matic, transponovat, hodnota
#else
 loc=glGetUniformLocationARB(program,NAME);
 if (loc!=-1) glUniformMatrix4fvARB(loc,C,GL_FALSE,M);
#endif
}

// konstruktor *************************************************************
CDOFshaders::CDOFshaders()
{
 loopi(5) { sh[i].vertex=0; sh[i].fragment=0; sh[i].program=0; }
}

// destruktor **************************************************************
CDOFshaders::~CDOFshaders()
{
 cancel(); // zrusenie DOF shaderov
}

// inicializacia DOF shaderov **********************************************
int CDOFshaders::init(void)
{
 char *vp_src,*fp_src,*fname; // pre vytvorenie DOF efektu je pouzitych 5 shaderov
 const char *vp_files[5]={"p0.vp","p1.vp","p2.vp","p3.vp","p4.vp"};
 const char *fp_files[5]={"p0.fp","p1.fp","p2.fp","p3.fp","p4.fp"};

 loopi(5) {
  // nazov suboru s kodom shadera
  fname=(char *) malloc(sizeof(char)*(strlen("shaders/dof/")+strlen(vp_files[i])+1));
  strcpy(fname,"shaders/dof/"); strcat(fname,vp_files[i]);
  // vytvorenie shader objektu (kontajnera)
  sh[i].vertex=glCreateShader(GL_VERTEX_SHADER);
  vp_src=TextFileRead(fname); // kod vertex shadera
  if (vp_src==NULL) {
   printf("Nemozno nacitat kod vertex shadera!\n");
   DELETEA(vp_src); DELETEA(fp_src); DELETEA(fname); return(1);
  }
  printf("vertex shader[%d]: %s\n",i,fname); DELETEA(fname);
  // kod shadera
  glShaderSource(sh[i].vertex,1,(const char**) &vp_src,NULL);
  // kompilacia shadera
  glCompileShader(sh[i].vertex);
  glGetShaderiv(sh[i].vertex,GL_INFO_LOG_LENGTH,&log_length);
  if (log_length>0 && log_length<1000) {
   glGetShaderInfoLog(sh[i].vertex,log_length,NULL,log);
   printf("vertex_shader_log: %s",log);
  }

  // nazov suboru s kodom shadera
  fname=(char *) malloc(sizeof(char)*(strlen("shaders/dof/")+strlen(fp_files[i])+1));
  strcpy(fname,"shaders/dof/"); strcat(fname,fp_files[i]);
  // vytvorenie shader objektu (kontajnera)
  sh[i].fragment=glCreateShader(GL_FRAGMENT_SHADER);
  fp_src=TextFileRead(fname); // kod fragment shadera
  if (fp_src==NULL) {
   printf("Nemozno nacitat kod fragment shadera!\n");
   DELETEA(vp_src); DELETEA(fp_src); DELETEA(fname); return(1);
  }
  printf("fragment shader[%d]: %s\n",i,fname); DELETEA(fname);
  // kod shadera
  glShaderSource(sh[i].fragment,1,(const char**) &fp_src,NULL);
  // kompilacia shadera
  glCompileShader(sh[i].fragment);
  glGetShaderiv(sh[i].fragment,GL_INFO_LOG_LENGTH,&log_length);
  if (log_length>0 && log_length<1000) {
   glGetShaderInfoLog(sh[i].fragment,log_length,NULL,log);
   printf("fragment_shader_log: %s",log);
  }

  // vytvorenie program objektu (kontajnera)
  sh[i].program=glCreateProgram();
  // pripojenie shaderov
  glAttachShader(sh[i].program,sh[i].vertex);
  glAttachShader(sh[i].program,sh[i].fragment);
  // zlozenie programu
  glLinkProgram(sh[i].program);
  glGetProgramiv(sh[i].program,GL_INFO_LOG_LENGTH,&log_length);
  if (log_length>0 && log_length<1000) {
   glGetProgramInfoLog(sh[i].program,log_length,NULL,log);
   printf("program_log: %s\n",log);
  }

  // uvolnenie pamate vyhradenej pre kod shaderov
  DELETEA(vp_src); DELETEA(fp_src);
 }

 return(0);
}

// zrusenie DOF shaderov ***************************************************
void CDOFshaders::cancel(void)
{
 loopi(5) {
  if (sh[i].vertex && sh[i].fragment && sh[i].program) {
   // odpojenie shaderov
   glDetachShader(sh[i].program,sh[i].vertex);
   glDetachShader(sh[i].program,sh[i].fragment);
   // zrusenie shader objektov
   glDeleteShader(sh[i].vertex);
   glDeleteShader(sh[i].fragment);
   // zrusenie program objektu
   glDeleteProgram(sh[i].program);
  }
 }
}

// zapnutie DOF shaderov ***************************************************
void CDOFshaders::on(int P)
{
 // pouzitie programu
 glUseProgram(sh[P].program);
}

// vypnutie DOF shaderov ***************************************************
void CDOFshaders::off(void)
{
 // nepouzitie programu
 glUseProgram(0);
}

// nastavenie atributov ****************************************************
void CDOFshaders::bindattriblocation(int P,const GLuint I,const char *NAME)
{
 // pripojenie atributu
 // atribut NAME programu program sa spoji s generickym atributom I
 // program musi byt znovu linkovany
 glBindAttribLocation(sh[P].program,I,NAME);
 glLinkProgram(sh[P].program);
}

// nastavenie uniform premennych *******************************************
void CDOFshaders::uniformI(int P,const char *NAME,const int I)
{
 // umiestnenie premennej
 loc=glGetUniformLocation(sh[P].program,NAME);
 // nastavenie premennej
 if (loc!=-1) glUniform1i(loc,I);
}
void CDOFshaders::uniform1F(int P,const char *NAME,const float F)
{
 // umiestnenie premennej
 loc=glGetUniformLocation(sh[P].program,NAME);
 // nastavenie premennej
 if (loc!=-1) glUniform1f(loc,F);
}
void CDOFshaders::uniformmatrixFV(int P,const char *NAME,const GLfloat *M)
{
 // umiestnenie matice
 loc=glGetUniformLocation(sh[P].program,NAME);
 // nastavenie matice
 if (loc!=-1) glUniformMatrix4fv(loc,1,GL_FALSE,M); // umiestnenie, 1 matica, transponovat, hodnota
}
void CDOFshaders::uniform4FV(int P,const char *NAME,const GLfloat *V)
{
 // umiestnenie premennej
 loc=glGetUniformLocation(sh[P].program,NAME);
 // nastavenie premennej
 if (loc!=-1) glUniform4fv(loc,4,V); // umiestnenie, 4 prvky, hodnota
}
