#version 130 // GLSL verzia

uniform sampler2D depth_tex_unit;  // hlbky
uniform sampler2D albedo_tex_unit; // albedo
uniform sampler2D normal_tex_unit; // normaly

// 4 matice s polohami a difuznymi komponentami 16-tich svetelnych zdrojov
uniform mat4x4 light_pos[4];
uniform mat4x4 light_dif[4];

// scale bias projection, inverse scale bias projection
uniform mat4x4 sbproj,isbproj;

// vstup z vertex shadera
in vec2 uv; // texturovacie suradnice
in vec4 ambient;

float depth; // hlbka fragmentu
vec4 N;      // normala v EYE space
vec4 FP,LP;  // 3D poloha fragmentu a poloha svetelneho zdroja
vec4 L;      // vektor VERTEX -> LIGHT v EYE space

mat4x4 m,n;  // pomocne matice
float NdotL; // pomocna premenna

void main(void)
{
 vec4 color,amb,light;
 float lng,lng2,att;

 // hlbka fragmentu ulozena v 2D texture
 depth=texture(depth_tex_unit,uv).r;

 // farba fragmentu ulozena v 2D texture
 color=texture(albedo_tex_unit,uv);

 // ambient
 amb=color*ambient;

 if (depth<1.0) {
  light=vec4(0.0);

  // normalizovana normala v EYE space ulozena v 2D texture v intervale <0;1>
  N=texture(normal_tex_unit,uv)*2.0-1.0;

  // vypocet 3D polohy fragmentu v EYE space zo znamej hlbky a 2D polohy
  // poloha v CLIP space = SCALE-BIAS * PROJECTION * poloha v EYE space
  // poloha v EYE space = iPROJECTION * iSCALE-BIAS * poloha v CLIP space
  FP=isbproj*vec4(uv,depth,1.0); FP.xyz/=FP.w;

  // pre 4 matice s polohami a difuznymi komponentami 16-tich svetelnych zdrojov
  for (int i=0;i<4;i++) {
   // vyber konkretnej matice - poloha a difuzny komponent
   m=light_pos[i]; n=light_dif[i];
   // pre polohy 4-roch svetelnych zdrojov
   for (int j=0;j<4;j++) {
    // 3D poloha sveteleneho zdroja v EYE space
    LP=m[j];
    // vektor FRAGMENT -> LIGHT v EYE space
    L=LP-FP; lng2=dot(L,L); lng=sqrt(lng2);
    // intenzita difuznej zlozky v zavislosti od polohy svetelneho zdroja
    NdotL=max(dot(N,L/lng),0.0);
    // constant+linear+quadratic attenuation
    att=gl_LightSource[0].constantAttenuation+
        gl_LightSource[0].linearAttenuation*lng+
        gl_LightSource[0].quadraticAttenuation*lng2;
    // diffuse
    light+=(n[j]*NdotL)/(0.5+att*exp(lng2-15.0));
   }
  }
 }

 // vysledna farba
 gl_FragData[0]=amb+color*light;
}
