#version 130 // GLSL verzia

#extension GL_EXT_texture_array : enable

uniform sampler2D base_tex_unit;          // zakladna textura
uniform sampler2DArrayShadow sm_tex_unit; // shadowmap pole textur

in vec3 nor;
in vec2 uv;
in vec3 dir_to_light,dir_to_eye;
in float dist;
in vec4 ambient,diffuse,specular;

// texturovacie suradnice pre shadowmap, resp. poloha fragmentu
// vypocitana pomocou i-tej transformacnej matice svetelneho zdroja
in vec4 sm_uv[5];

vec4 sm_uvw; // sm_uv perspective divided
float illumination; // osvetlenie fragmentu

vec3 zero=vec3(0.0);
vec3 one=vec3(1.0);

void main()
{
 vec4 color,amb,dif=vec4(0.0),spe=vec4(0.0);
 vec3 N,L,E,RV;
 float NdotL,att=0.0,spot;
 int i;

 // farba z textury
 color=texture(base_tex_unit,uv);

 // ambient
 amb=color*ambient;

 // normalizacia interpolovanych udajov
 N=normalize(nor);
 L=normalize(dir_to_light);

 NdotL=max(dot(N,L),0.0);
 if (NdotL>0.0) {
  spot=dot(gl_LightSource[0].spotDirection,-L);
  if (spot>gl_LightSource[0].spotCosCutoff) {
   // utlm
   att=pow(spot,gl_LightSource[0].spotExponent)/
       (gl_LightSource[0].constantAttenuation+
        gl_LightSource[0].linearAttenuation*dist+
        gl_LightSource[0].quadraticAttenuation*dist*dist);

   // diffuse
   dif=NdotL*diffuse;

   // specular
   E=normalize(dir_to_eye);
   RV=normalize(-reflect(L,N)); // reflect vector
   spe=pow(max(dot(E,RV),0.0),gl_FrontMaterial.shininess)*specular;
  }
 }

 // osvetlnenie fragmentu
 illumination=0.0;

 // pre vsetkych 5 vrstiev shadowmap
 for (i=0;i<5;i++) {
  // poloha fragmentu vypocitana pomocou i-tej transformacnej matice svetelneho zdroja
  sm_uvw=sm_uv[i]/sm_uv[i].q; // perspective divide

  // ohranicenie texture lookup do frustum objemu svetelneho zdroja
  if (all(greaterThan(sm_uvw.stp,zero)) && all(lessThan(sm_uvw.stp,one))) {

   // vzdialenost (hlbka) aktualneho fragmentu od polohy svetelneho zdroja
   // "Z = sm_uvw.p = prvok q vektora = R pre GL_COMPARE_R_TO_TEXTURE"
   // je porovnavana s hlbkou ulozenou v shadowmap (do shadowmap sa renderuju odvratene
   // steny objektov), ak je hlbka fragmentu mensia, resp. rovnaka, fragment je osvetleny,
   // v opacnom pripade je fragment v tieni => hodnota illumination zostane nulova
   illumination+=texture(sm_tex_unit,vec4(sm_uvw.st,i,sm_uvw.p));

  }
 }

 // aritmeticky priemer 5-tich texture lookup
 illumination/=5.0;

 // vysledna farba
 //gl_FragData[0]=vec4(illumination,illumination,illumination,1.0);
 gl_FragData[0]=amb+att*(dif+spe)*illumination;
}
